from collections import OrderedDict

###################################################################################
#  词法分析器                                                                     #
###################################################################################

INTEGER = 'INTEGER'  # 整数类型
REAL = 'REAL'  # 实数类型
INTEGER_CONST = 'INTEGER_CONST'  # 整数（因子）
REAL_CONST = 'REAL_CONST'  # 实数（因子）
PLUS = 'PLUS'  # 加
MINUS = 'MINUS'  # 减
MUL = 'MUL'  # 乘
INTEGER_DIV = 'INTEGER_DIV'  # 整数除法
FLOAT_DIV = 'FLOAT_DIV'  # 浮点数除法
LPAREN = 'LPAREN'  # 左括号
RPAREN = 'RPAREN'  # 右括号
ID = 'ID'  # 变量名称
ASSIGN = 'ASSIGN'  # 赋值符号
BEGIN = 'BEGIN'  # 开始标记
END = 'END'  # 结束标记
SEMI = 'SEMI'  # 分号
DOT = 'DOT'  # 点（程序结束符）
PROGRAM = 'PROGRAM'  # 程序
VAR = 'VAR'  # 变量声明标记
COLON = 'COLON'  # 冒号
COMMA = 'COMMA'  # 逗号
EOF = 'EOF'  # 结束符号


class Token:
    def __init__(self, value_type, value):
        self.value_type = value_type
        self.value = value

    def __str__(self):
        return 'Token({value_type},{value})'.format(value_type=self.value_type, value=self.value)

    def __repr__(self):
        return self.__str__()


RESERVED_KEYWORDS = {  # 保留字
    'PROGRAM': Token('PROGRAM', 'PROGRAM'),
    'VAR': Token('VAR', 'VAR'),
    'DIV': Token('INTEGER_DIV', 'DIV'),
    'INTEGER': Token('INTEGER', 'INTEGER'),
    'REAL': Token('REAL', 'REAL'),
    'BEGIN': Token('BEGIN', 'BEGIN'),
    'END': Token('END', 'END'),
}

class Lexer:
    def __init__(self, text):
        self.text = text
        self.position = 0
        self.current_char = self.text[self.position]

    def error(self):
        raise Exception('警告：错误的输入内容！')

    def number(self):
        result = ''
        while self.current_char is not None and self.current_char.isdigit():
            result += self.current_char
            self.advance()
            if self.current_char == '.':
                result += self.current_char
                self.advance()
                while self.current_char is not None and self.current_char.isdigit():
                    result += self.current_char
                    self.advance()
                return Token(REAL_CONST, float(result))
        return Token(INTEGER_CONST, int(result))

    def skip_comment(self):
        while self.current_char != '}':
            self.advance()
        self.advance()

    def _id(self):
        result = ''
        if self.current_char == '_':
            result = '_'
            self.advance()
        while self.current_char is not None and self.current_char.isalnum():
            result += self.current_char
            self.advance()
        token = RESERVED_KEYWORDS.get(result.upper(), Token('ID', result.lower()))
        return token

    def peek(self):
        pos = self.position + 1
        if pos >= len(self.text):
            return None
        else:
            return self.text[pos]

    def skip_whitespace(self):
        while self.current_char is not None and self.current_char.isspace():
            self.advance()

    def advance(self):
        self.position += 1
        if self.position >= len(self.text):
            self.current_char = None
        else:
            self.current_char = self.text[self.position]

    def get_next_token(self):
        while self.current_char is not None:
            if self.current_char == '{':
                self.advance()
                self.skip_comment()
                continue
            if self.current_char.isspace():
                self.skip_whitespace()
                continue
            if self.current_char.isalpha() or self.current_char == '_':
                return self._id()
            if self.current_char == ':' and self.peek() == '=':
                self.advance()
                self.advance()
                return Token(ASSIGN, ':=')
            if self.current_char == ';':
                self.advance()
                return Token(SEMI, ';')
            if self.current_char == '.':
                self.advance()
                return Token(DOT, '.')
            if self.current_char.isdigit():
                return self.number()
            if self.current_char == ':':
                self.advance()
                return Token(COLON, ':')
            if self.current_char == ',':
                self.advance()
                return Token(COMMA, ',')
            if self.current_char == '+':
                self.advance()
                return Token(PLUS, '+')
            if self.current_char == '-':
                self.advance()
                return Token(MINUS, '-')
            if self.current_char == '*':
                self.advance()
                return Token(MUL, '*')
            if self.current_char == '/':
                self.advance()
                return Token(FLOAT_DIV, '/')
            if self.current_char == '(':
                self.advance()
                return Token(LPAREN, '(')
            if self.current_char == ')':
                self.advance()
                return Token(RPAREN, ')')
            self.error()
        return Token(EOF, None)


###################################################################################
#  语法分析器                                                                     #
###################################################################################
class AST:
    pass


class Program(AST):
    def __init__(self, name, block):
        self.name = name
        self.block = block


class Block(AST):
    def __init__(self, declarations, compound_statement):
        self.declarations = declarations
        self.compound_statement = compound_statement


class VarDecl(AST):
    def __init__(self, var_node, type_node):
        self.var_node = var_node
        self.type_node = type_node


class Type(AST):
    def __init__(self, token):
        self.token = token
        self.name = token.value


class Compound(AST):
    def __init__(self):
        self.children = []


class Assign(AST):
    def __init__(self, left, operator, right):
        self.left = left
        self.token = self.operator = operator
        self.right = right


class Variable(AST):
    def __init__(self, token):
        self.token = token
        self.name = token.value


class NoOperator(AST):
    pass


class BinaryOperator(AST):
    def __init__(self, left, operator, right):
        self.left = left
        self.token = self.operator = operator
        self.right = right


class Num(AST):
    def __init__(self, token):
        self.token = token
        self.value = token.value


class UnaryOperator(AST):
    def __init__(self, operator, expr):
        self.token = self.operator = operator
        self.expr = expr


class Parser:
    def __init__(self, lexer):
        self.lexer = lexer
        self.current_token = lexer.get_next_token()

    def error(self):
        raise Exception('警告：错误的输入内容！')

    def eat(self, value_type):
        if self.current_token.value_type == value_type:
            self.current_token = self.lexer.get_next_token()
        else:
            self.error()

    def program(self):
        self.eat(PROGRAM)
        var_node = self.variable()
        program_name = var_node.name
        self.eat(SEMI)
        block_node = self.block()
        node = Program(program_name, block_node)
        self.eat(DOT)
        return node

    def block(self):
        declarations = self.declarations()
        compound_statement = self.compound_statement()
        node = Block(declarations, compound_statement)
        return node

    def declarations(self):
        declarations = []
        if self.current_token.value_type == VAR:
            self.eat(VAR)
            while self.current_token.value_type == ID:
                declarations.extend(self.variable_declaration())
                self.eat(SEMI)
        return declarations

    def variable_declaration(self):
        var_nodes = [Variable(self.current_token)]
        self.eat(ID)
        while self.current_token.value_type == COMMA:
            self.eat(COMMA)
            var_nodes.append(Variable(self.current_token))
            self.eat(ID)
        self.eat(COLON)
        type_node = self.type_spec()
        var_declarations = [VarDecl(var_node, type_node) for var_node in var_nodes]
        return var_declarations

    def type_spec(self):
        token = self.current_token
        if token.value_type == INTEGER:
            self.eat(INTEGER)
        else:
            self.eat(REAL)
        node = Type(token)
        return node

    def compound_statement(self):
        self.eat(BEGIN)
        nodes = self.statement_list()
        self.eat(END)

        root = Compound()
        root.children = nodes
        return root

    def statement_list(self):
        node = self.statement()
        nodes = [node]
        while self.current_token.value_type == SEMI:
            self.eat(SEMI)
            nodes.append(self.statement())
        if self.current_token.value_type == ID:
            self.error()
        return nodes

    def statement(self):
        if self.current_token.value_type == BEGIN:
            node = self.compound_statement()
        elif self.current_token.value_type == ID:
            node = self.assignment_statement()
        else:
            node = self.empty()
        return node

    def assignment_statement(self):
        left = self.variable()
        token = self.current_token
        self.eat(ASSIGN)
        right = self.expr()
        node = Assign(left, token, right)
        return node

    def empty(self):
        return NoOperator()

    def variable(self):
        node = Variable(self.current_token)
        self.eat(ID)
        return node

    def factor(self):
        current_token = self.current_token

        if current_token.value_type == PLUS:
            self.eat(PLUS)
            node = UnaryOperator(current_token, self.factor())
            return node
        elif current_token.value_type == MINUS:
            self.eat(MINUS)
            node = UnaryOperator(current_token, self.factor())
            return node
        elif current_token.value_type == INTEGER_CONST:
            self.eat(INTEGER_CONST)
            return Num(current_token)
        elif current_token.value_type == REAL_CONST:
            self.eat(REAL_CONST)
            return Num(current_token)
        elif current_token.value_type == LPAREN:
            self.eat(LPAREN)
            node = self.expr()
            self.eat(RPAREN)
            return node
        else:
            node = self.variable()
            return node

    def term(self):
        node = self.factor()
        while self.current_token.value_type in (MUL, INTEGER_DIV, FLOAT_DIV):
            token = self.current_token
            if token.value_type == MUL:
                self.eat(MUL)
            elif token.value_type == INTEGER_DIV:
                self.eat(INTEGER_DIV)
            elif token.value_type == FLOAT_DIV:
                self.eat(FLOAT_DIV)
            node = BinaryOperator(node, token, self.factor())
        return node

    def expr(self):
        node = self.term()
        while self.current_token.value_type in (PLUS, MINUS):
            token = self.current_token
            if token.value_type == PLUS:
                self.eat(PLUS)
            elif token.value_type == MINUS:
                self.eat(MINUS)
            node = BinaryOperator(node, token, self.term())
        return node

    def parser(self):
        node = self.program()
        if self.current_token.value_type != EOF:
            self.error()
        return node


###################################################################################
#  AST访问器                                                                      #
###################################################################################
class NodeVisitor:
    def visit(self, node):
        method_name = 'visit_' + type(node).__name__
        visitor = getattr(self, method_name, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node):
        raise Exception(f'未找到“visit_{type(node).__name__}()”方法！')


###################################################################################
#  符号和符号表                                                                   #
###################################################################################

class Symbol:  # 添加符号类
    def __init__(self, name, symbol_type=None):
        self.name = name  # 符号名称
        self.symbol_type = symbol_type  # 符号类型


class BuiltinTypeSymbol(Symbol):  # 添加内置类型符号类
    def __init__(self, name):
        super().__init__(name)  # 调用基类构造函数初始化

    def __str__(self):
        return self.name  # 返回符号名称

    __repr__ = __str__


class VarSymbol(Symbol):  # 添加变量符号类
    def __init__(self, name, symbol_type):
        super().__init__(name, symbol_type)  # 调用基类构造函数初始化

    def __str__(self):
        return f'<{self.name}:{self.symbol_type}>'  # 返回符号信息

    __repr__ = __str__


class SymbolTable:  # 添加符号表类
    def __init__(self):
        self._symbols = OrderedDict()  # 存储符号的有序字典
        self.__init_builtins()  # 初始化内置类型符号

    def __init_builtins(self):  # 添加初始化内置类型符号的方法
        self.define(BuiltinTypeSymbol('INTEGER'))  # 定义整数类型符号
        self.define(BuiltinTypeSymbol('REAL'))  # 定义实数类型符号

    def __str__(self):
        return f'Symbols:{[value for value in self._symbols.values()]}'  # 返回符号列表信息

    __repr__ = __str__

    def define(self, symbol):  # 添加定义符号的方法
        print(f'定义：{symbol}')
        self._symbols[symbol.name] = symbol  # 符号添加到字典

    def lookup(self, name):  # 添加查询符号的方法
        print(f'查询：{name}')
        return self._symbols.get(name)  # 返回符号对象


class SymbolTableBuilder(NodeVisitor):  # 添加符号表生成器类
    def __init__(self):  # 初始化
        self.symbol_table = SymbolTable()  # 创建符号表对象

    def visit_Program(self, node):  # 添加访问程序节点的方法
        self.visit(node.block)  # 访问块节点（块中可能包含符号）

    def visit_Block(self, node):  # 添加访问块节点的方法
        for declaration in node.declarations:  # 遍历声明节点
            self.visit(declaration)  # 访问声明语句节点（声明语句中可能包含符号）
        self.visit(node.compound_statement)  # 访问复合语句节点

    def visit_VarDecl(self, node):  # 添加访问变量声明节点方法

        type_name = node.type_node.name  # 获取变量类型名称
        symbol_type = self.symbol_table.lookup(type_name)  # 获取变量类型
        var_name = node.var_node.name  # 获取变量名称
        var_symbol = VarSymbol(var_name, symbol_type)  # 创建变量符号对象
        self.symbol_table.define(var_symbol)  # 变量符号对象添加到符号表

    def visit_Compound(self, node):  # 添加访问复合语句节点的方法
        for child in node.children:  # 遍历语句节点列表
            self.visit(child)  # 访问语句节点（语句中可能包含符号）

    def visit_NoOperator(self, node):  # 添加访问空语句的节点
        pass  # 不做任何处理（空语句之后的语句可能包含符号，需要跳过空语句）

    def visit_BinaryOperator(self, node):  # 添加访问二元运算符方法（二元运算中可能包含符号）
        self.left = self.visit(node.left)  # 访问左侧节点
        self.right = self.visit(node.right)  # 访问右侧节点

    def visit_Num(self, node):  # 添加访问数字的方法
        pass  # 不做任何处理（数字节点之后的内容可能包含符号，需要跳过数字，因为不进行解释，也无需返回数字的值）

    def visit_UnaryOperator(self, node):  # 添加访问一元运算符方法
        self.visit(node.expr)  # 访问表达式节点（表达式中可能包含符号）

    def visit_Assign(self, node):  # 添加访问赋值节点的方法
        var_name = node.left.name  # 获取变量名称
        var_symbol = self.symbol_table.lookup(var_name)  # 查询变量符号
        if var_symbol is None:  # 如果变量未声明
            raise NameError(f'错误的标识符：{repr(var_name)}')  # 抛出名称错误异常
        self.visit(node.right)  # 访问右侧节点（赋值语句右侧可能包含符号）

    def visit_Variable(self, node):  # 添加访问变量节点的方法
        var_name = node.name  # 获取变量名称
        var_symbol = self.symbol_table.lookup(var_name)  # 查询变量符号
        if var_symbol is None:  # 如果变量未声明
            raise NameError(f'错误的标识符：{repr(var_name)}')  # 抛出名称错误异常


###################################################################################
#  解释器                                                                         #
###################################################################################
class Interpreter(NodeVisitor):  # 修改解释器
    def __init__(self, tree):  # 修改构造方法
        self.tree = tree  # 获取参数AST
        self.GLOBAL_MEMORY = OrderedDict()  # 创建全局存储字典

    def visit_Program(self, node):
        self.visit(node.block)

    def visit_Block(self, node):
        for declaration in node.declarations:
            self.visit(declaration)
        self.visit(node.compound_statement)

    def visit_VarDecl(self, node):
        pass

    def visit_Type(self, node):
        pass

    def visit_Compound(self, node):
        for child in node.children:
            self.visit(child)

    def visit_Assign(self, node):
        var_name = node.left.name
        self.GLOBAL_MEMORY[var_name] = self.visit(node.right)

    def visit_Variable(self, node):
        var_name = node.name
        value = self.GLOBAL_MEMORY.get(var_name)
        if value is None:
            raise NameError(f'错误的标识符：{repr(var_name)}')
        else:
            return value

    def visit_NoOperator(self, node):
        pass

    def visit_BinaryOperator(self, node):
        if node.operator.value_type == PLUS:
            return self.visit(node.left) + self.visit(node.right)
        elif node.operator.value_type == MINUS:
            return self.visit(node.left) - self.visit(node.right)
        elif node.operator.value_type == MUL:
            return self.visit(node.left) * self.visit(node.right)
        elif node.operator.value_type == INTEGER_DIV:
            return self.visit(node.left) // self.visit(node.right)
        elif node.operator.value_type == FLOAT_DIV:
            return self.visit(node.left) / self.visit(node.right)

    def visit_Num(self, node):
        return node.value

    def visit_UnaryOperator(self, node):
        if node.operator.value_type == PLUS:
            return +self.visit(node.expr)
        if node.operator.value_type == MINUS:
            return -self.visit(node.expr)

    def interpret(self):  # 修改解释方法
        tree = self.tree  # 获取AST
        if tree is None:  # 如果AST不存在
            return ''  # 返回空字符串
        return self.visit(tree)  # 否则访问AST


def main():
    text = '''
        PROGRAM Part11;
        VAR
           number : INTEGER;
           a, b   : INTEGER;
           y      : REAL;
        
        BEGIN {Part11}
           number := 2;
           a := number ;
           b := 10 * a + 10 * number DIV 4;
           y := 20 / 7 + 3.14
        END.  {Part11}
    '''
    lexer = Lexer(text)
    parser = Parser(lexer)
    tree = parser.parser()
    symtab_builder = SymbolTableBuilder()
    symtab_builder.visit(tree)
    print('符号表中的内容：\n', symtab_builder.symbol_table)
    print('-----------------------------------------------')
    print('开始解释代码...')
    interpreter = Interpreter(tree)
    interpreter.interpret()
    print('全局存储中的内容：')
    for k, v in interpreter.GLOBAL_MEMORY.items():
        print(f'{k}：{v}')


if __name__ == '__main__':
    main()
